//***************************************************************************
//  File........: adc.c
//  Target(s)...: ATmega328p
//  Compiler....: AVR-GCC
//  Description.: AVR ADC routines
//***************************************************************************

#include <stdint.h> // Integer-Datentypen 8-64bit
#include <avr/io.h> // Pins und Interrupt-Tabelle -> siehe auch iom328p.h
#include "main.h"  // CPU-Frequenz, sbi, cbi
#include "adc.h" // ADC-Funktionen
#include "bcd.h" // BCD-Umwandlungen fuer Ausgabe
#include "usart.h" // USART-Funktionen
#include "delay.h" // Wartezeiten
//#include <avr/pgmspace.h>
//#include "pgmspacehlp.h"
//#include "LCD_functions.h"
//#include "timer0.h"
//#include "dataflash.h"


//extern char gLogging;

/*****************************************************************************
*   Function name : ADC_init
*   Returns :       None
*   Parameters :    char ref
*                   1 : AREF=1.1V
*                   0 : AREF=VCC
*   Purpose :       Initialize ADC with	ADC-channel 0 Pin 23
*                   bereitet ADC-Wandlung vor
*****************************************************************************/
void ADC_init(char ref)
{
    //unsigned int i;  // war char. Nur geht for-Schleife dann in die Hose bei >255

    /*
    USART_Tx_msg(PSTR(" ADC-Init: "));  // Hilfsausgabe
    USART_Tx_msg(PSTR("ADCSRB vor: ")); // Hilfsausgabe
    USART_Tx_Int(ADCSRB);	    // Hilfsausgabe
    USART_Tx_msg(PSTR(" ADCSRA vor: ")); // Hilfsausgabe
    USART_Tx_Int(ADCSRA);	    // Hilfsausgabe
    */
    ADCSRB = (0<<ADTS2) |(0<<ADTS1) |(0<<ADTS0);  // ADC free running mode

    ADCSRA = (1<<ADEN) | (0<<ADIE) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);
    // schaltet ADC aktiv, keine ADC-Interrupts, Prescaler 128 -> 16MHz/128=125kHz
    // da ADC aktiv kann/muss Referenz auch aktiv sein
    // ADC-Wandlung recht langsam, koennte doppelt so schnell sein.

    // ggf. Powerreduction bit PRADC nutzen im PowerReductionRegister PRR.
    // ADC-Takt knnte abgeschaltet werden fr weniger Stromaufnahme.
    // ggf. erst abschalten wenn deaktiviert ist.

    /*
    USART_Tx_msg(PSTR("ADCSRB nach: ")); // Hilfsausgabe
    USART_Tx_Int(ADCSRB);	    // Hilfsausgabe
    USART_Tx_msg(PSTR(" ADCSRA nach: ")); // Hilfsausgabe
    USART_Tx_Int(ADCSRA);	    // Hilfsausgabe
    */
    /*
    USART_Tx_msg(PSTR("ADC-Init:")); // Hilfsausgabe
    USART_Tx_msg(PSTR("input:")); // Hilfsausgabe
    USART_Tx_Int(input);	   // Hilfsausgabe
    USART_Tx_msg(PSTR("ADMUX:")); // Hilfsausgabe
    USART_Tx_Int(ADMUX);	   // Hilfsausgabe
    */

    if (ref == 1){ // wenn interne Referenz 1.1V verwendet werden soll
		      // sonst Referenz = VCC.

	/*
	USART_Tx_msg(PSTR("Entladung bei Uin-Messung:")); // Hilfsausgabe
	USART_Tx_msg(PSTR("ADMUX:")); // Hilfsausgabe
	USART_Tx_Int(ADMUX);	       // Hilfsausgabe
	*/

	ADMUX = (1<<REFS1) | (1<<REFS0) | (0<<ADLAR) | (0<<MUX3) | (0<<MUX2) | (0<<MUX1) | (0<<MUX0);
	// AD-Wandlung ADC0, interne 1.1V Referenz

	// warte bis 0.1uF an VREF auf 1.1V aufgeladen.
	delay_us(500);

    }
    else {   //	also externe Referenz VCC verwendet.

	ADMUX = (0<<REFS1) | (1<<REFS0) | (0<<ADLAR) | (0<<MUX3) | (0<<MUX2) | (0<<MUX1) | (0<<MUX0);
	// AD-Wandlung ADC0, externe VCC-Referenz

	// warte bis 0.1uF an VREF auf 5V aufgeladen.
	delay_us(500);

    }

    // dummy read to allow references to stabilize
    ADCSRA = (1<<ADEN) | (1<<ADSC) | (0<<ADATE)	| (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);
					// set ADC prescaler to 16MHz / 128 = 125kHz
    while(ADCSRA & (1<<ADSC));		// wait for conversion done, ADSC flag to zero
}


/*****************************************************************************
*   Function name : ADC_read
*   Returns :	    int ADC value
*   Parameters :    None
*   Purpose :	    Do a Analog to Digital Conversion
*		    bestimmt einen ADC-Wert entsprechend der Einstellung in ADMUX
*****************************************************************************/
int ADC_read(void)
{
    int ADC_temp;

    //do a dummy readout first. Wozu denn? Um 1.1V-Referenz wieder hochzufahren? !!
//    ADCSRA |=(1<<ADEN)| (1<<ADSC);	  // do single conversion, ADC einschalten
					// Referenz braucht ggf. ca. 70us Zeit um hochzufahren !
					// ein delay sinnvoll von >70us
//    delay_us(500);

//    while(ADCSRA & (1<<ADSC));	  // wait for conversion done, ADSC flag to zero

    //USART_Tx_msg(PSTR("ADC_read: ")); // Hilfsausgabe

    // do the ADC conversion
	ADCSRA |= (1<<ADSC);		// do single conversion
	while(ADCSRA & (1<<ADSC));	// wait for conversion done, ADSC flag to zero

	ADC_temp = ADCL;			// read ADCL register. Immer erst low auslesen!
	ADC_temp += (ADCH << 8);	// read ADCH register. Obere 2 bit !

	/*
	USART_Tx_msg(PSTR(" ADC_read: ")); // Hilfsausgabe
	USART_Tx_Int(ADC_temp);		// Hilfsausgabe
	*/

    return ADC_temp;
}


/*****************************************************************************
*   Function name : ADC_read_mean
*   Returns :	    int ADC average value after 8 reads
*   Parameters :    None
*   Purpose :	    Do a Analog to Digital Conversion
*		    bestimmt einen ADC-Wert entsprechend der Einstellung in ADMUX
*		    mittelt Wert 8x fuer bessere Genauigkeit
*****************************************************************************/
int ADC_read_mean(void)
{
    char i;
    int ADC_temp;
    int ADCr = 0;

    //do a dummy readout first. Wozu denn? Um 1.1V-Referenz wieder hochzufahren? !!
//    ADCSRA |=(1<<ADEN)| (1<<ADSC);	  // do single conversion, ADC einschalten
					// Referenz braucht ggf. ca. 70us Zeit um hochzufahren !
					// ein delay sinnvoll von >70us
//    delay_us(500);

//    while(ADCSRA & (1<<ADSC));	  // wait for conversion done, ADSC flag to zero


    //USART_Tx_msg(PSTR("ADC_read: ")); // Hilfsausgabe

    // do ADC conversion 8 times for better accuracy
    for(i=0;i<8;i++){

	ADCSRA |= (1<<ADSC);		// do single conversion
	while(ADCSRA & (1<<ADSC));	// wait for conversion done, ADSC flag to zero

	ADC_temp = ADCL;			// read ADCL register. Immer erst low auslesen!
	ADC_temp += (ADCH << 8);	// read ADCH register. Obere 2 bit !

	/*
	USART_Tx_msg(PSTR(" ADC_read 8: ")); // Hilfsausgabe
	USART_Tx_Int(ADC_temp);               // Hilfsausgabe
	*/

	ADCr +=	ADC_temp;           // accumulate result (8 samples) for later averaging
    }

    // average the 8 samples
    ADCr = ADCr >> 3;

    return ADCr;
}


/*****************************************************************************
*   Function name : ADC_periphery
*   Returns :	    None
*   Parameters :    None
*   Purpose :	    Calculates the Temperature/Voltage/Ligth from the ADC_read
*                   and puts it out on the LCD.
*****************************************************************************/
/* void ADC_periphery(void)
{
    int	ADCresult = 0;
    int	inverse	= 0;

    ADCresult=ADC_read();	      // read ADC Value
				      // ADC auslesen dessen Kanal in ADMUX festgelegt ist

    //process ADC value	accordingly
    switch ((ADMUX & 7)){	      // schaut	was im ADMUX-Register steht -> Kanal
				      // ADMUX verundet	mit 00000111 ->	Kanal 0-7
	case 0:      // =0 falls es der Temperatursensor ist
	    ADC2Temp(ADCresult);// LCD-Ausgabe Temperatur
	    break;
	case 1:	      // =1 falls es die Spannung an Uin ist
	    ADC2Volt(ADCresult);// LCD-Ausgabe
	    break;
	case 2:	      // =2 falls es der Lichtsensor ist
	    //invert for light readings	more light = higher reading
	    inverse = 1023-ADCresult; // damit viel Licht mit hheren Werten ausgegeben	wird
	    ADC2RAW(ADCresult);   // LCD-Ausgabe Lichtsensor-Rohwerte 0-1023 = 0-A23
	    break;
	case 3:	      // =4 falls es der Stromsensor ist
	    ADC2Curr(ADCresult);// LCD-Ausgabe
	    break;
	default:		      // falls es weder der Temp-, Uin- oder Lichtsensor ist
	    ADC2RAWAUX(ADCresult,(ADMUX &0x07)); // konvertiert	ADC-Rohwerte fuer Ausgabe LCD,UART
    }
}
*/


/*****************************************************************************
*   Function name : ADC2RAWAUX
*   Returns :       nothing
*   Parameters :    char log (char to log to flash else	print to screen)
*                   int ADCResult (ADC reading to convert)
*   Purpose :       Convert an ADC reading into a raw Reading
*		    gibt ADC-Werte roh von einem der 4 JTAG-Pins aus
*		    Ausgabe auf LCD und serieller Schnittstelle
*****************************************************************************/
/*void ADC2RAWAUX(int ADCresult)
{
    //	int Temp_int;
    int Temp;
    char TL;
    char TH;
    char THH;

    Temp = INT2BCD3(ADCresult);

    TL = (Temp & 0x0F) + '0';
    if(TL > '9')	// if the hex-value is over 9, add 7 in	order to go
	TL += 7;	// jump	to the character in the	ASCII-table

    //TH = ((Temp >>= 4)& 0x0F)	+ '0';
    TH = ((Temp	>>4)& 0x0F) + '0';
    if(TH > '9')	// if the hex-value is over 9, add 7 in	order to go
	TH += 7;	// jump	to the character in the	ASCII-table

    //THH = ((Temp >>= 4)& 0x0F) + '0';
    THH	= ((Temp >>4)& 0x0F) + '0';
    if(THH > '9')	 // if the hex-value is	over 9,	add 7 in order to go
	THH += 7;	 // jump to the	character in the ASCII-table

	USART_Tx(THH);
	USART_Tx(TH);
	USART_Tx(TL);
	USART_Tx(' ');

	//LCD_putc(0, 'T');	// print x,y or	z
	LCD_putc(0, '0');	// print 1,2,3 or 4
	LCD_putc(1, ' ');
	LCD_putc(2, ' ');
	LCD_putc(3, THH);
	LCD_putc(4, TH);
	LCD_putc(5, TL);
	LCD_putc(6, '\0');

}
*/

/*****************************************************************************
*   Function name : ADC2Volt
*   Returns :	    nothing
*   Parameters :    int ADCResult (ADC reading to convert)
*   Purpose :	    Convert an ADC reading into a Voltage
*		    wandelt ADC-Wert in Spannung um
*		    Spannungsteiler 1.5MOhm zu 300kOhm auf Board -> Teiler 6. Max. 6.6V messbar
*		    erweiterter	Spannungsteiler durch externe 2.7MOhm. Max. 16.5V messbar
*		    wandelt alternativ ADC-Wert in Strom um
*		    Stromsensoren wie ACS 752SCA-050 einsetzbar fuer +-50A, bis +-75A nichtlinear
*		    Spannungsteiler 1.5MOhm zu 300kOhm auf Board -> Teiler 6. Max. 6.6V messbar
*		    Spannungsteiler 300k zu 300k auf Board -> Teiler 2. Max. 2.2V messbar
*		    Spannungsteiler 600k zu 300k auf Board -> Teiler 3. Max. 3.3V messbar
*		    Stromsensor ACS 752SCA-050 -> 0 bis -50A (Entladung) -> 2.5V bis 0.5V 1V=25A
*					       -> 0 bis 50A (Ladung) -> 2.5V bis 4.5V
*					       -> 0 bis 25A (Ladung) -> 2.5V bis 3.5V 1V=25A
*					       -> 0 bis 20A (Ladung) -> 2.5V bis 3.3V 0.1V=2.5A 0.2V=5A
*		    Stromaufloesung bei 6.6V-Bereich: 6.6V/1023=6.45mV Aufloesung mit 1mV=0.025A
*					       -> 1 LSB=0.025A*6.45=0.161A
*		    Stromaufloesung bei 3.3V-Bereich: 3.3V/1023=3.2mV -> 1 LSB=0.025A*3.2=0.08A=80mA
*****************************************************************************/
void ADC2Volt(int ADCresult)
{
    //	int ADCresult_temp = 0;
    char Temp;

    char TH;  // damit auch 16V noch darstellbar sind
    char TL;
    char VH;
    char VL;
    char negativ =0;

    float V_ADC;
    char VoltageHB;
    char VoltageLB;

// ****	nur eine define-Zeile aktiv lassen ******

    //V_ADC = ADCresult * 0.6452f;    // aus 1023 wird 660 -> mult. mit 660/1023=0.6452
    //V_ADC = ADCresult * 0.006452f;  // aus 1023 wird 6.60V -> mult. mit 6.60/1023=0.006452
    V_ADC = ADCresult * 0.006354f;    // aus 1023 wird 6.5V -> mult. mit 6.5/1023=0.006354 (Modul 2)

				    // 10bit=1024 Stufen, Vref=1.1V
				    // warum der Aufwand mit float?
				    // ginge auch mit Integer zu lsen.
				    // unsigned int geht bis 65535.
				    // ADC kann 1023 sein=6.60V
				    // V_ADC=1023*1.10/1024  von 0-1.0989V
				    // V_ADC=1023*1.10/1023  von 0-1.1V scheint besser
				    // *110/1024 leicht zu rechnen
				    // *110*6/1024 auch denkbar
				    // *110*6/1024 wrde jedoch berlauf geben -> long sinnvoll
				    // *660/1024 bedeutet nun 2 Stellen hinter Komma
				    // eine INT2BCD4-Wandlung wrde es auch tun.
				    // so ist es wohl anschaulicher und vom define 1.1f einfacher
				    // zeitlich krzer wre vielleicht der andere Weg
				    // 1023*660/1024 659 ergibt mit /100 das VoltageHB=6
				    //			 ergibt mit 659-6*100=VoltageLB=59

//    V_ADC = V_ADC *6;  // Multiply by 6 because of the voltage division
				         // Ua an 300k, Ue an 1500k+300k=1800k 1800/300=6

/* Umrechnung fuer Ausgabe */

    if (V_ADC <0){ // vermeide Probleme mit negativen Werten
		V_ADC =-V_ADC;
		negativ= 1;
    }

    VoltageHB = V_ADC;        // Store high-byte. 04.12 -> 04	ins HighByte
    V_ADC = V_ADC-VoltageHB;  // 4.12-4=0.12
    VoltageLB =V_ADC* 100;    // Store low-byte. 0.12*100 -> 12 ins Low-Byte

    Temp = CHAR2BCD2(VoltageHB);    // Convert from char to BCD

    TH = (Temp >> 4) + '0';	    // 04 -> 0 -> 30 hinzu gibt dies als 0 aus
    TL = (Temp & 0x0F) + '0';	    // 04 -> 4 -> 30 hinzu gibt dies als 4 aus

    Temp = CHAR2BCD2(VoltageLB);    // Convert from char to bin

    VH = (Temp >> 4) + '0';
    VL = (Temp & 0x0F) + '0';

    // Testausgaben
    /*USART_Tx_msg(PSTR(" ADCresult ")); // Hilfsausgabe
      USART_Tx_Int(ADCresult);		 // Hilfsausgabe
      USART_Tx_msg(PSTR(" HB "));
      USART_Tx_Int(VoltageHB);
      USART_Tx_msg(PSTR(" LB "));
      USART_Tx_msg(PSTR(" "));
      USART_Tx_Int(VoltageLB); */
      // Ende Testausgaben

     if (TH == '0'){ // fuehrende Null da -> ignorieren
     }
     else {
	    USART_Tx(TH);  // gebe TH aus wenn es keine Null ist
     }
     USART_Tx(TL);
     USART_Tx(',');
     USART_Tx(VH);
     USART_Tx(VL);

/*
 if (negativ==1){
		LCD_putc(0, '-');
	}
	else {
	    LCD_putc(0, ' ');
	}
	if (TH == '0')
		TH = ' ';   // unterdruecke fuehrende Null

	LCD_putc(1, TH); // gebe auch TH aus
	LCD_putc(2, TL);

	LCD_putc(3, 'v');
	LCD_putc(4, VH);
	LCD_putc(5, VL);
	LCD_putc(6, '\0');
*/
}


/*****************************************************************************
*   Function name : ADC2RAW
*   Returns :       nothing
*   Parameters :    char log (char to log to flash else	print to screen)
*                   int ADCResult (ADC reading to convert)
*   Purpose :       Convert an ADC reading into	a Raw Reading
*                   Ausgabe ADC-Rohwert
*****************************************************************************/
/*
void ADC2RAW(unsigned int ADCresult)
{
    //int Temp_int;
    int Temp;
    //char i = 0;
    char TL;
    char TH;
    char THH;


    Temp = INT2BCD3(ADCresult);

    TL = (Temp & 0x0F) + '0';  // nur die unteren 4 bit, als hex ausgeben
    if(TL > '9')	// if hex-value is over 9, add 7 in order to go
	TL += 7;	// jump to character in ASCII-table

    TH = ((Temp >>= 4)& 0x0F) + '0'; // mittlere 4 bit nach rechts schieben, Rest wegmaskieren
    if(TH > '9')	// if hex-value is over 9, add 7
	TH += 7;	// jump to character in ASCII-table

    THH	= ((Temp >>= 4)& 0x0F) + '0'; // obere 4 bit nach rechts schieben, Rest wegmaskieren
    if(THH > '9')	 // if hex-value is over 9, add 7
	THH += 7;	 // jump to character in ASCII-table

    if (THH>'9'){ // adjust to print 10 rather than A. Statt A23 nun 1023 ausgeben
	USART_Tx('1');
	THH = '0';
     }
     USART_Tx(THH);
     USART_Tx(TH);
     USART_Tx(TL);
}
*/


/*
        ; -- AD-Wandler-Beispiel fuer ATmega80535
        out     ADMUX,null      ; verwende Eingang AD0
        sbi     ADCSRA,ADSC     ; startet ADC im free run

wait_adc:       ; warte bis der ADC den ersten Wert auch hat
        sbis    ADCSRA,ADIF     ; teste ob Wandlung fertig ist
        rjmp    wait_adc        ; nein, dann warte noch

                ; lese dann ADC ein
        in      temp,ADCL       ; ADC lesen: erst lo, dann hi !! beide noetig
        in      temp3,ADCH      ; ADC hi lesen

*/
