//***************************************************************************
//  File........: usart.c
//  Target(s)...: ATmega328p
//  Compiler....: AVR-GCC
//  Description.: AVR USART routines
//***************************************************************************

#include <stdint.h> // Integer-Datentypen 8-64bit
#include <avr/io.h> // Pins und Interrupt-Tabelle -> siehe auch iom328p.h
#include "main.h"  // CPU-Frequenz, sbi, cbi
#include "usart.h" // USART-Funktionen
//#include "adc.h" // ADC-Funktionen
#include "bcd.h" // BCD-Umwandlungen fuer Ausgabe
//#include "RTC.h"
//#include "alarm.h"
//#include "pgmspacehlp.h"
//#include "LCD_functions.h"
//#include "LCD_driver.h"
#include <avr/interrupt.h>
#include <avr/pgmspace.h>


/*****************************************************************************
*   Function name : USART_Init
*   Returns :       None
*   Parameters :    unsigned int baudrate
*   Purpose :       Initialize the USART
*****************************************************************************/
void USART_init(unsigned int baudrate)
{
    // Set baud rate
    UBRR0H = (unsigned char)(baudrate>>8);
    UBRR0L = (unsigned char)baudrate;

    // Enable 2x speed
    UCSR0A = (1<<U2X0);

    // Enable receiver and transmitter
    UCSR0B = (1<<RXCIE0)|(0<<TXCIE0)|(0<<UDRIE0)|(1<<RXEN0)|(1<<TXEN0); // mit Interrupt

	// Async. mode, kein parity, 1 Stop-Bit, 8bit, async
	UCSR0C = (0<<UMSEL01)|(0<<UMSEL00)|(0<<UPM01)|(0<<UPM00)|(0<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00)|(0<<UCPOL0);

}


/*****************************************************************************
*   Function name : USART_Tx
*   Returns :       None
*   Parameters :    char data: byte to send
*   Purpose :       Send one byte through the USART
*                   ein Byte seriell schicken
*****************************************************************************/
void USART_Tx(char data)
{
    while (!(UCSR0A & (1<<UDRE0)));  // schauen ob Senden ok ist
    UDR0 = data;   // Datenbyte abschicken, wenn erlaubt
}


/*****************************************************************************
*   Function name : USART_Rx
*   Returns :       char: byte received
*   Parameters :    None
*   Purpose :       Receives one byte from the USART
*                   ein Byte seriell empfangen
*****************************************************************************/
char USART_Rx(void)
{
    while (!(UCSR0A & (1<<RXC0)));
    return UDR0;  // Datenbyte empfangen, wenn erlaubt
}


/*****************************************************************************
*   Function name : USART_Tx_hex
*   Returns :       None
*   Parameters :    char data: byte to send
*   Purpose :       Sends one byte through the USART as a two char Hex number
*                   Byte seriell schicken als 2 Hex-Ziffern
*                   z.B. 0x00 =0, 0x08 =8, 0x0a =9 0xff =255
*                   der GCC wird mit -funsigned-char genutzt -> unsigned char !
*
*                   z.B.  0 = 0000 0000 als 00 ausgeben -> 30 und 30 senden (zwei ASCII-Nullen)
*                         9 = 0000 1001 als 09 ausgeben -> 30 und 39 senden
*                        10 = 0000 1010 als 0A ausgeben -> 30 und 'A'=41 senden
*                        15 = 0000 1111 als 0F ausgeben
*                       255 = 1111 1111 als FF ausgeben
*****************************************************************************/
void USART_Tx_hex(char data)
{
    unsigned char low, high;

    high = data >> 4;  // 8 bit-Wert 4 bit nach rechts schieben, also erst high-Teil bearbeiten
    high += '0';       // 30 addieren setzt also die linken 4 bit auf 0011=3
    if (high > '9') high += 7;  // wenn die Ziffer >9 ist (z.B.10=3A) dann 7 addieren (hier 41=A)

    low = data & 0x0f; // unterer Teil der 8bit, also obere 4bit auf 0 setzen
    low += '0';        // wieder 30 addieren
    if (low > '9') low += 7; // wie gehabt bei >39 (z.B 3A) 7 addieren (z.B. 3B,3C,3D,3E,3F,40,41=A)

    USART_Tx(high);  // schicke erst High-Byte, dann Low-Byte
    USART_Tx(low);

}


/*****************************************************************************
*   Function name : USART_Tx_byte
*   Returns :       None
*   Parameters :    char data: byte to send
*   Purpose :       Sends one byte through the USART as a 2 digit decimal number
*                   ein Byte als 2 Dezimalziffern senden (nur 0-99 zulssig)
*                     0 (=0x00) als 0 0 senden
*                     1 (=0x01) als 0 1 senden
*                     9 (=0x09) als 0 9 senden
*                    10 (=0x0A) als 1 0 senden
*                    15 (=0x0F) als 1 5 senden
*                    16 (=0x10) als 1 6 senden
*                    99 (=0x63) als 9 9 senden
*                   100 (=0x64) als A 0 senden (ab hier wird es unsinnig, daher verboten !)
*                   101 (=0x65) als A 1 senden?
*                   109 (=0x6D) als A 9 senden?
*                   110 (=0x6E) als B 0 senden?
*                   120         als C 0 senden?
*                   130         als D 0 senden?
*                   140         als E 0 senden?
*                   150         als F 0 senden?
*                   151         als F 1 senden?
*                   159         als F 9 senden?
*                   255 (=0xFF) als ???
*
*****************************************************************************/
void USART_Tx_byte(char data)
{
    unsigned char temp;

	temp = CHAR2BCD2(data);
	USART_Tx_hex(temp);
}

/*****************************************************************************
*   Function name : USART_Tx_int
*   Returns :       None
*   Parameters :    unsigned int data: 10bit value to send
*   Purpose :       Sends value through the USART as a 4 digit decimal number
*                   Wert als 4 Dezimalziffern senden (nur 0-1023 zulssig)
*                     0 (=0x00) als 0 0 0 0 senden
*                     1 (=0x01) als 0 0 0 1 senden
*                     9 (=0x09) als 0 0 0 9 senden
*                    10 (=0x0A) als 0 0 1 0 senden
*                    15 (=0x0F) als 0 0 1 5 senden
*                    16 (=0x10) als 0 0 1 6 senden
*                    99 (=0x63) als 0 0 9 9 senden
*                   100 (=0x64) als 0 1 0 0 senden
*                  1000         als 1 0 0 0 senden
*                  1023         als 1 0 2 3 senden
*                  9999         als 9 9 9 9 senden
*                 10000         als A 0 0 0 senden? (ab hier wird es unsinnig, daher verboten !)
*                 65535         als F F F F senden?
*****************************************************************************/
void USART_Tx_int(unsigned int data)
{
	unsigned long temp;
	unsigned char tmp;
	unsigned char ziff=0; // fuehrende Null da? fuer Unterdrueckung
	//data=65535;   // Test
	signed char i;  // muss signed sein, damit for-Schleife beendet wird mit <0 !

	temp = INT2BCD5(data);  // xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx 2 Integer a 16bit
	                        //   0    0    0    0    0    0    0    0
	                        //   0    0    0    6    5    5    3    5

	//USART_Tx ((( (unsigned char) (temp>>16) )&0x0f)+'0'); // soll eine ASCII 6 ausgeben
	//USART_Tx ((( (unsigned char) (temp>>12) )&0x0f)+'0');
	//USART_Tx ((( (unsigned char) (temp>>8) )&0x0f)+'0');
	//USART_Tx ((( (unsigned char) (temp>>4) )&0x0f)+'0');
	//USART_Tx ((( (unsigned char) (temp>>0) )&0x0f)+'0');

	for (i=16; i>=0; i-=4)
	{
		// i=16 0x00065535 -> 0x00000006 -> 0x06 -> 0x06 -> 36 ausgeben
		// i=12 0x00065535 -> 0x00000065 -> 0x65 -> 0x05 -> 35 ausgeben
		// i=8  0x00065535 -> 0x00000655 -> 0x55 -> 0x05 -> 35 ausgeben
		// i=4  0x00065535 -> 0x00006553 -> 0x53 -> 0x03 -> 33 ausgeben
		// i=0  0x00065535 -> 0x00065535 -> 0x35 -> 0x05 -> 35 ausgeben

		tmp= (((unsigned char) (temp>>i)) &0x0f)+'0'; // einzelnes Zeichen: 30 addieren fr ASCII-Ausgabe

		if (tmp=='0') { // eine Ziffer = Null da?
			if (ziff==0) { // ist es eine fuehrende 0?
			    if (i==0) { // ist die letzte Ziffer eine 0?
				    USART_Tx (tmp); // diese Null ausgeben
				}
				else { // nein, also unterdruecken
					USART_Tx (' '); // fuehrende Null als Space ausgeben
				}
		    }
		    else {
		    	USART_Tx (tmp); // Nullen in der Zahl als Nullen ausgeben
			}
		}
		else { // eine Ziffer ungleich Null da
			ziff=1;
			USART_Tx (tmp); // Ziffer 0-9 als ASCII ausgeben
		}
	}
}



/****************************************************************************
*	Function Name :	ISR_USART_RX
*	Returns : 		None
*	Parameters :	None
*	Purpose :		Interrupt service routine for Data received from UART
*					Receives one byte of data from the USART. Determines
*					command char received and execute code as required.
* vgl. WinAVR-20100110/doc/avr-libc/avr-libc-user-manual/group__avr__interrupts.html
* Eintraege zu ATmega328P
*****************************************************************************/
ISR(USART_RX_vect)
{
	char receive;
	//unsigned int i;
	//unsigned int wert;
	//unsigned int rambuffer[500]; // Werte im RAM

	receive = USART_Rx(); // get command character from UART receive register
	USART_Tx (receive);   // empfangenes Zeichen zur Bestaetigung ruecksenden
	USART_Tx ('\n');      // neue Zeile

	//is it a number for ADC channel
	if (receive=='S'){  // ADC starten?
	    cli(); // ADC abarbeiten -> weitere Interrupts solange sperren

      //  USART_Tx_msg(PSTR("ADC-Wandlung ins RAM starten:\n"));  // Meldung ausgeben

		// ADC-Wandlung, Schreiben ins RAM und UART-Ausgabe abarbeiten
      //  for (i=0; i<500;i++){ // 0-499 sind 500 Werte
		//   wert = ADC_read(); // einen ADC-Rohwert holen (0-1023)
		 //  rambuffer[i]=wert; // Rohwert ins RAM schreiben
		   //USART_Tx_Int(wert); // Rohwert testweise ausgeben
		   //USART_Tx ('\n');  // neue Zeile nach jedem Wert
		//} // Rohwerte stehen nun im RAM

		// ADC-Werte aus RAM formatiert ausgeben
		//USART_Tx_msg(PSTR("ADC-Werte aus RAM ausgeben:\n"));  // Meldung ausgeben
		//for (i=0; i<500;i++){ // 0-499 sind 500 Rohwerte
		  // USART_Tx_Int(i); // index ausgeben (0-499)
		  // USART_Tx ('\t');  // Tab zur Trennung
		  // USART_Tx_Int(rambuffer[i]); // Wert ausgeben (0-1023)
		  // USART_Tx ('\n');  // neue Zeile nach jedem Wert
		//} // Werte aus RAM seriell ausgegeben
		//USART_Tx_msg(PSTR("ADC-Werte fertig ausgegeben!\n"));  // Meldung ausgeben

		sei(); // Interrupts wieder freigeben
	}
	else{ // anderes Zeichen - ignorieren

	}

}


/*****************************************************************************
*   Function name : USART_Tx_msg
*   Returns :       nothing
*   Parameters :    pointer to program memory string
*   Purpose :       Sends a string out on the UART
*****************************************************************************/
void USART_Tx_msg (const char *s)
{
	char c;
	while ((c=pgm_read_byte(s++)) ){	// get next char, bis /0
		USART_Tx(c);	//send char
	}
}


/*****************************************************************************
*   Function name : USART_sendADC
*   Returns :       nothing
*   Parameters :    char channel to read
*   Purpose :       Read from an ADC and send it out via the UART
*****************************************************************************/
/*
void USART_sendADC(char channel)

{
	int logdata;

	ADC_init(channel);					// set up adc
	logdata=ADC_read();					// read adc

	//ADC2RAW(logdata);
}
*/



/*****************************************************************************
*   Function name : USART_getValue
*   Returns :       int - 0 for unsuccessful / 1 for successful
*   Parameters :    int
*   Purpose :       Takes an input from user via Uart and does basic error check
*****************************************************************************/
/*
char USART_getValue(char *value, char lowerbound, char upperbound)
{
	char retval;						//function return value
	char temp;
	retval = 1;							//set retval to 1 as default (successful)
	*value = 0;							//initialise the parameter

	// print out the limits for the input
	USART_Tx_byte(lowerbound);
	USART_Tx('-');
	USART_Tx_byte(upperbound);
	USART_Tx(' ');

	temp = USART_Rx();					//get first character sent
	*value = ((char)(temp - '0')) * 10;	//convert char from Ascii to Hex and multiply by 10

	temp = USART_Rx();					//get second character sent
	*value += ((char)(temp - '0'));		//convert char from Ascii to Hex and add to yearValue

	if ((*value < lowerbound) | (*value > upperbound)){
		//check the value provided by user is valid
		//if not then increment error counter return 0

		retval = 0;						//set retval to 0 signifying unsuccessful operation
		USART_Tx_msg(PSTR("ERR:"));

		return retval;
	}


	return retval;
}
*/
